"""
A slightly better approach to reloading modules and function than the standard
[`importlib.reload()`](https://docs.python.org/3/library/importlib.html#importlib.reload)
function. The functions in this module are for interactive use in a Python REPL.

* `reload_module(module)`: reloads the given module and all its parent modules
* `reload_function(func)`: reloads and returns the given function

!!! note
    It might be possible that after the module or function has been reloaded that an extra
    import is needed to import the proper module attributes in your namespace.

Module dependency

When you make a change in a module or function that you are not calling directly, but call
through another function from another module, you need to reload the module where you made
the change and, after that, reload the function that calls that module.

Example:
    * module `x.a` contains function `func_a`
    * module `x.b` contains function `func_b` which calls `func_a`

    when you make a change in `func_a` you have to reload the module `x.a` and then reload
    the function `func_b`:

    ```python
    from x.b import func_b
    func_b()
    ```

    now make some changes in `func_a`, then to make those changes known in your session:

    ```python
    reload_module('x.a')
    func_b = reload_function(func_b)
    func_b()  # will show the changes done in func_a
    ```
"""

import importlib
import itertools
import typing
import types

import rich

from egse.exceptions import Abort


def reload_module(module: typing.Union[types.ModuleType, str], walk: bool = True):
    """
    Reloads the given module and all its parent modules. The modules will be reloaded starting
    from their top-level module. Reloading the 'egse.hexapod.symetry.puna' module will reload
    'egse', 'egse.hexapod', 'egse.hexapod.symetry', and 'egse.hexapod.symetry.puna' in that order.

    Note:
        If you pass the module argument as a module, make sure the top level module is
        imported in your session, or you will get a NameError. To prevent this (if you don't want
        to import the top-level module, pass the module as a string.

    Example:
        ```python
        import egse
        reload_module(egse.system)
        ```
        or
        ```python
        reload_module('egse.system')
        ```

    Args:
        module: The module that needs to be reloaded
        walk: walk up the module hierarchy and import all modules [default=True]

    """
    full_module_name = module.__name__ if isinstance(module, types.ModuleType) else module

    module_names = (
        itertools.accumulate(full_module_name.split("."), lambda x, y: f"{x}.{y}") if walk else [full_module_name]
    )

    for module_name in module_names:
        try:
            module = importlib.import_module(module_name)
            importlib.reload(module)
        except (Exception,) as exc:
            rich.print(f"[red]Failed to reload {module_name}[/red], {exc=}")


def reload_function(func: types.FunctionType) -> types.FunctionType:
    """
    Reloads and returns the given function. In order for this to work, you should catch the
    return value to replace the given function with its reloaded counterpart.

    This will also work if you import the function again instead of catching it.

    Note:
        that this mechanism doesn't work for functions that were defined in the `__main__` module.

    Example:
        ```python
        func = reload_function(func)
        ```
        or
        ```python
        reload_function(func)
        from egse.some_module import func
        ```

    Args:
        func: the function that needs to be reloaded

    Returns:
        The reloaded function.

    Raises:
        Abort: when the function is not the proper type or when the function is defined
            in the `__main__` module.
    """

    # Why do I raise an Abort instead of just printing a message and returning?
    #
    # The function is usually called catching the return value and replacing the same function with
    # its reloaded counterpart. If we would just return (None) when an error occurs, the original
    # function will be overwritten with None. Raising an Exception leaves the original as it was
    # before the call.

    if not isinstance(func, types.FunctionType):
        raise Abort(f"The 'func' argument shall be a function, not {type(func)}")

    module_name = func.__module__
    func_name = func.__name__

    if module_name == "__main__":
        raise Abort("Cannot reload a function that is defined in the __main__ module.")

    reload_module(module_name)

    module = __import__(module_name, fromlist=[func_name])
    return getattr(module, func_name)



__all__ = [
    "GlobalState",
]

import abc
import textwrap

from egse.setup import Setup
from egse.setup import load_setup
from egse.setup import setup_ctx


class StateError(Exception):
    pass


class UnknownStateError(StateError):
    pass


class IllegalStateTransition(StateError):
    pass


class NotImplementedTransition(StateError):
    pass


class ConnectionStateInterface(abc.ABC):
    """
    A class used to enforce the implementation of the _connection_ interface
    to model the state of a (network) connection.

    Subclasses only need to implement those methods that are applicable to
    their state.

    """

    # This class is to enforce the implementation of the interface on both the
    # model, i.e. the Proxy, and the State class. At the same time, it will allow
    # the State subclasses to implement only those methods that are applicable
    # in their state.

    @abc.abstractmethod
    def connect(self, proxy):
        pass

    @abc.abstractmethod
    def disconnect(self, proxy):
        pass

    @abc.abstractmethod
    def reconnect(self, proxy):
        pass


class _GlobalState:
    """
    This class implements global state that is shared between instances of this class.
    """

    def __call__(self, *args, **kwargs):
        return self

    @property
    def setup(self) -> Setup | None:
        """
        Returns the currently active Setup for this process.

        NOTE: The returned Setup is not necessarily the currently active Setup in the
              configuration manager. That is only true of your process monitors the
              notifications from the configuration manager and updates the Setup when this
              is changed in/by the configuration manager. Use the `GlobalState.load_setup()`
              method to force loading the Setup from the configuration manager.

        Returns:
            The currently active Setup or None.
        """
        return setup_ctx.get()

    def load_setup(self) -> Setup | None:
        """
        Loads the currently active Setup from the Configuration manager. The current Setup is the Setup
        that is defined and loaded in the Configuration manager. When the configuration manager is not
        reachable, None will be returned, a warning will be logged, and the global state setup will be cleared.

        Since the GlobalState should reflect the configuration of the test, it can only load the current
        Setup from the configuration manager. If you need to work with different Setups, work with the `Setup`
        class and the Configuration Manager directly.

        Returns:
            The currently active Setup or None.
        """
        return load_setup()


GlobalState = _GlobalState()

if __name__ == "__main__":
    import rich

    rich.print(
        textwrap.dedent(
            f"""\
            GlobalState info:
              Setup loaded: {GlobalState.setup.get_id() if GlobalState.setup else "[orange3]no Setup loaded[/]"}
            """
        )
    )



"""
The Settings class handles user and configuration settings that are provided in
a [`YAML`](http://yaml.org) file.

The idea is that settings are grouped by components or any arbitrary grouping that makes sense for
the application or for the user. Settings are also modular and provided by each package by means
of entry-points. The Settings class can read from different YAML files.

By default, settings are loaded from a file called `settings.yaml`, but this can be changed in the entry-point
definition.

The yaml configuration files are provided as entry points by the packages that specified an entry-point group
'_cgse.settings_' in the `pyproject.toml`. The Settings dictionary (attrdict) is constructed from the configuration
YAML files from each of the packages. Settings can be overwritten by the next package configuration file. So,
make sure the group names in each package configuration file are unique.

The YAML file is read and the configuration parameters for the given group are
available as instance variables of the returned class.

The intended use is as follows:

```python
from egse.settings import Settings

dsi_settings = Settings.load("DSI")

if 0x000C <= dsi_settings.RMAP_BASE_ADDRESS <= 0x00FF:
    ...  # do something here
else:
    raise RMAPError("Attempt to access outside the RMAP memory map.")
```

The above code reads the settings from the default YAML file for a group called `DSI`.
The settings will then be available as variables of the returned class, in this case
`dsi_settings`. The returned class is and behaves also like a dictionary, so you can
check if a configuration parameter is defined like this:

```python
if "DSI_FEE_IP_ADDRESS" not in dsi_settings:
    # define the IP address of the DSI
```
The YAML section for the above code looks like this:

```text
DSI:

    # DSI Specific Settings

    DSI_FEE_IP_ADDRESS  10.33.178.144   # IP address of the DSI EtherSpaceLink interface
    LINK_SPEED:                   100   # SpW link speed used for both up- and downlink

    # RMAP Specific Settings

    RMAP_BASE_ADDRESS:     0x00000000   # The start of the RMAP memory map managed by the FEE
    RMAP_MEMORY_SIZE:            4096   # The size of the RMAP memory map managed by the FEE
```

When you want to read settings from another YAML file, specify the `filename=` keyword.
If that file is located at a specific location, also use the `location=` keyword.

    my_settings = Settings.load(filename="user.yaml", location="/Users/JohnDoe")

The above code will read the YAML file from the given location and not from the entry-points.

---

"""

import re
from pathlib import Path
from typing import Any

import yaml  # This module is provided by the pip package PyYaml - pip install pyyaml

from egse.env import bool_env
from egse.env import get_local_settings_env_name
from egse.env import get_local_settings_path
from egse.log import logger
from egse.system import attrdict
from egse.system import recursive_dict_update

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")


class SettingsError(Exception):
    """A settings-specific error."""

    pass


# Fix the problem: YAML loads 5e-6 as string and not a number
# https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number

SAFE_LOADER = yaml.SafeLoader
SAFE_LOADER.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
             [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)


def load_settings_file(path: Path, filename: str, force: bool = False) -> attrdict:
    """
    Loads the YAML configuration file that is located at `path / filename`.

    Args:
        path (PATH): the folder where the YAML file is located
        filename (str): the name of the YAML configuration file
        force (bool): force reloading, i.e. don't use the cached information

    Raises:
        SettingsError: when the configuration file doesn't exist or cannot be found or \
        when there was an error reading the configuration file.

    Returns:
         A dictionary (attrdict) with all the settings from the given file.

    Note:
        in case of an empty configuration file, and empty dictionary \
        is returned and a warning message is issued.
    """
    try:
        yaml_document = read_configuration_file(path / filename, force=force)
        settings = attrdict({name: value for name, value in yaml_document.items()})
    except FileNotFoundError as exc:
        raise SettingsError(f"The Settings YAML file '{filename}' is not found at {path!s}. ") from exc

    if not settings:
        logger.warning(
            f"The Settings YAML file '{filename}' at {path!s} is empty. "
            f"No local settings were loaded, an empty dictionary is returned."
        )

    return settings


def load_global_settings(entry_point: str = "cgse.settings", force: bool = False) -> attrdict:
    """
    Loads the settings that are defined by the given entry_point. The entry-points are defined in the
    `pyproject.toml` files of the packages that export their global settings.

    Args:
         entry_point (str): the name of the entry-point group [default: 'cgse.settings']
         force (bool): force reloading the settings, i.e. ignore the cache

    Returns:
        A dictionary (attrdict) containing a collection of all the settings exported by the packages \
        through the given entry-point.

    """
    from egse.plugin import get_file_infos

    ep_settings = get_file_infos(entry_point)

    global_settings = attrdict(label="Settings")

    for ep_name, (path, filename) in ep_settings.items():
        settings = load_settings_file(path, filename, force)
        recursive_dict_update(global_settings, settings)

    return global_settings


def load_local_settings(force: bool = False) -> attrdict:
    """
    Loads the local settings file that is defined from the environment variable *PROJECT*_LOCAL_SETTINGS (where
    *PROJECT* is the name of your project, defined in the environment variable of the same name).

    This function might return an empty dictionary when

      - the local settings YAML file is empty
      - the local settings environment variable is not defined.

    in both cases a warning message is logged.

    Raises:
        SettingsError: when the local settings YAML file is not found. Check the *PROJECT*_LOCAL_SETTINGS \
        environment variable.

    Returns:
        A dictionary (attrdict) with all local settings.

    """
    local_settings = attrdict()

    local_settings_path = get_local_settings_path()
    if VERBOSE_DEBUG:
        logger.debug(f"{get_local_settings_env_name()=}")
        logger.debug(f"{local_settings_path=}")

    if local_settings_path:
        path = Path(local_settings_path).expanduser()
        local_settings = load_settings_file(path.parent, path.name, force)

    return local_settings


def read_configuration_file(filename: Path, *, force=False) -> dict:
    """
    Read the YAML input configuration file. The configuration file is only read
    once and memoized as load optimization.

    Args:
        filename (Path): the fully qualified filename of the YAML file
        force (bool): force reloading the file, even when it was memoized

    Raises:
        SettingsError: when there was an error reading the YAML file.

    Returns:
        a dictionary containing all the configuration settings from the YAML file.
    """
    filename = str(filename)

    if force or not Settings.is_memoized(filename):
        if VERBOSE_DEBUG:
            logger.debug(f"Parsing YAML configuration file {filename}.")

        with open(filename, "r") as stream:
            try:
                yaml_document = yaml.load(stream, Loader=SAFE_LOADER)
            except yaml.YAMLError as exc:
                logger.error(exc)
                raise SettingsError(f"Error loading YAML document {filename}") from exc

        Settings.add_memoized(filename, yaml_document)

    return Settings.get_memoized(filename) or {}


class Settings:
    """
    The Settings class provides a load() method that loads configuration settings for a group
    into a dynamically created class as instance variables.
    """

    __memoized_yaml = {}  # Memoized settings yaml files
    __profile = False  # Used for profiling methods and functions

    LOG_FORMAT_DEFAULT = "%(levelname)s:%(module)s:%(lineno)d:%(message)s"
    LOG_FORMAT_FULL = "%(asctime)23s:%(levelname)8s:%(lineno)5d:%(name)-20s: %(message)s"
    LOG_FORMAT_THREAD = "%(asctime)23s:%(levelname)7s:%(lineno)5d:%(name)-20s(%(threadName)-15s): %(message)s"
    LOG_FORMAT_PROCESS = (
        "%(asctime)23s:%(levelname)7s:%(lineno)5d:%(name)20s.%(funcName)-31s(%(processName)-20s): %(message)s"
    )
    LOG_FORMAT_DATE = "%d/%m/%Y %H:%M:%S"

    @classmethod
    def get_memoized_locations(cls) -> list:
        return list(cls.__memoized_yaml.keys())

    @classmethod
    def is_memoized(cls, filename: str) -> bool:
        return filename in cls.__memoized_yaml

    @classmethod
    def add_memoized(cls, filename: str, yaml_document: Any):
        cls.__memoized_yaml[filename] = yaml_document

    @classmethod
    def get_memoized(cls, filename: str):
        return cls.__memoized_yaml.get(filename)

    @classmethod
    def clear_memoized(cls):
        cls.__memoized_yaml.clear()

    @classmethod
    def set_profiling(cls, flag):
        cls.__profile = flag

    @classmethod
    def profiling(cls):
        return cls.__profile

    @staticmethod
    def _load_all(
        entry_point: str = "cgse.settings", add_local_settings: bool = False, force: bool = False
    ) -> attrdict:
        """
        Loads all settings from all package with the entry point 'cgse.settings'
        """
        global_settings = load_global_settings(entry_point, force)

        # Load the LOCAL settings YAML file

        if add_local_settings:
            local_settings = load_local_settings(force)
            recursive_dict_update(global_settings, local_settings)

        return global_settings

    @staticmethod
    def _load_group(
        group_name: str, entry_point: str = "cgse.settings", add_local_settings: bool = False, force: bool = False
    ) -> attrdict:
        global_settings = load_global_settings(entry_point, force)

        group_settings = attrdict(label=group_name)

        if group_name in global_settings:
            group_settings = attrdict(
                {name: value for name, value in global_settings[group_name].items()}, label=group_name
            )

        if add_local_settings:
            local_settings = load_local_settings(force)
            if group_name in local_settings:
                recursive_dict_update(group_settings, local_settings[group_name])

        if not group_settings:
            raise SettingsError(f"Group name '{group_name}' is not defined in the global nor in the local settings.")

        return group_settings

    @staticmethod
    def _load_one(location: str, filename: str, force=False) -> attrdict:
        return load_settings_file(Path(location).expanduser(), filename, force)

    @classmethod
    def load(
        cls, group_name=None, filename="settings.yaml", location=None, *, add_local_settings=True, force=False
    ) -> attrdict:
        """
        Load the settings for the given group. When no group is provided, the
        complete configuration is returned.

        The Settings are loaded from entry-points that are defined in each of the
        packages that provide a Settings file.

        If a location is explicitly provided, the Settings will be loaded from that
        location, using the given filename or the default (which is settings.yaml).

        Args:
            group_name (str): the name of one of the main groups from the YAML file
            filename (str): the name of the YAML file to read [default=settings.yaml]
            location (str, Path): the path to the location of the YAML file
            force (bool): force reloading the file
            add_local_settings (bool): update the Settings with site specific local settings

        Returns:
            a dynamically created class with the configuration parameters as instance variables.

        Raises:
            SettingsError: when the group is not defined in the YAML file.
        """
        if group_name:
            return cls._load_group(group_name, add_local_settings=add_local_settings, force=force)
        elif location:
            return cls._load_one(location=location, filename=filename, force=force)
        else:
            return cls._load_all(add_local_settings=add_local_settings, force=force)

    @classmethod
    def to_string(cls):
        """
        Returns a simple string representation of the cached configuration of this Settings class.
        """
        memoized = cls.__memoized_yaml

        msg = ""
        for key in memoized.keys():
            msg += f"YAML file: {key}\n"
            for field in memoized[key].keys():
                length = 60
                line = str(memoized[key][field])
                trunc = line[:length]
                if len(line) > length:
                    trunc += " ..."
                msg += f"   {field}: {trunc}\n"

        return msg.rstrip()

    @classmethod
    def from_string(cls, yaml_string: str, group: str | None = None) -> attrdict:
        """
        Creates a Settings object from a YAML string.

        Args:
            yaml_string (str): the YAML configuration as a string

        Returns:
            An attribute dictionary with all the settings from the given string.
        """
        try:
            yaml_document = yaml.load(yaml_string, Loader=SAFE_LOADER)
            settings = attrdict({name: value for name, value in yaml_document.items()}, label="Settings")
        except yaml.YAMLError as exc:
            logger.error(exc)
            raise SettingsError("Error loading YAML document from string") from exc

        if not settings:
            logger.warning(
                "The Settings YAML string is empty. No local settings were loaded, an empty dictionary is returned."
            )

        if group:
            if group in settings:
                settings = attrdict({name: value for name, value in settings[group].items()}, label=group)
            else:
                raise SettingsError(f"Group name '{group}' is not defined in the provided YAML string.")

        return settings


def main(args: list | None = None):  # pragma: no cover
    # We provide convenience to inspect the settings by calling this module directly from Python.
    #
    # python -m egse.settings
    #
    # Use the '--help' option to see what your choices are.

    from egse.env import setup_env

    setup_env()

    import argparse

    parser = argparse.ArgumentParser(
        description=(
            f"Print out the default Settings, updated with local settings if the "
            f"{get_local_settings_env_name()} environment variable is set."
        ),
    )
    parser.add_argument("--local", action="store_true", help="print only the local settings.")
    parser.add_argument(
        "--global", action="store_true", help="print only the global settings, don't include local settings."
    )
    parser.add_argument("--group", help="print only settings for this group")
    args = parser.parse_args(args or [])

    # The following import will activate the pretty printing of the AttributeDict
    # through the __rich__ method.

    from rich import print

    if args.local:
        location = get_local_settings_path()
        if location:
            location = str(Path(location).expanduser().resolve())
            settings = Settings.load(filename=location)
            print(settings)
            print(f"Loaded from [purple]{location}.")
        else:
            print("[red]No local settings defined.")
    else:
        # if the global option is given we don't want to include local settings
        add_local_settings = False if vars(args)["global"] else True

        if args.group:
            settings = Settings.load(args.group, add_local_settings=add_local_settings)
        else:
            settings = Settings.load(add_local_settings=add_local_settings)
        print(settings)
        print("[blue]Memoized locations:")
        locations = Settings.get_memoized_locations()
        print([str(loc) for loc in locations])


def get_site_id() -> str:
    site = Settings.load("SITE")
    return site.ID


# ignore_m_warning('egse.settings')

if __name__ == "__main__":
    main()



"""
This module contains a number of convenience functions to work with bits, bytes and integers.
"""

from __future__ import annotations

import ctypes
from typing import Union


def extract_bits(value: int, start_position: int, num_bits: int) -> int:
    """
    Extracts a specified number of bits from an integer starting at a given position.

    Args:
        value (int): The input integer.
        start_position (int): The starting bit position (0-based index).
        num_bits (int): The number of bits to extract.

    Returns:
        int: The extracted bits as an integer.
    """
    # Create a mask with 'num_bits' set to 1
    mask = (1 << num_bits) - 1

    # Right shift the number by 'start_position' bits
    shifted_number = value >> start_position

    # Apply the mask to extract the desired bits
    extracted_bits = shifted_number & mask

    return extracted_bits


def set_bit(value: int, bit: int) -> int:
    """
    Set bit to 1 for the given value.

    Args:
        value (int): the integer value that needs a bit set or unset
        bit (int): the index of the bit to set/unset, starting from 0 at the LSB

    Returns:
        the changed value.
    """
    return value | (1 << bit)


def set_bits(value: int, bits: tuple[int, int]) -> int:
    """
    Set the given bits in value to 1.

    Args:
        value (int): the value where the given bits shall be changed
        bits (tuple): a tuple with start and stop bits

    Returns:
        the changed value.
    """
    for i in range(bits[0], bits[1]):
        value |= 1 << i
    return value


def clear_bit(value: int, bit: int) -> int:
    """
    Set bit to 0 for the given value.

    Args:
        value (int): the integer value that needs a bit set or unset
        bit (int): the index of the bit to set/unset, starting from 0 at the LSB

    Returns:
        the changed value.
    """
    return value & ~(1 << bit)


def clear_bits(value: int, bits: tuple[int, int]) -> int:
    """
    Set the given bits in value to 0.

    Args:
        value (int): the value where the given bits shall be changed
        bits (tuple): a tuple with start and stop bits

    Returns:
        the changed value
    """
    for i in range(bits[0], bits[1]):
        value &= ~(1 << i)
    return value


def toggle_bit(value: int, bit: int) -> int:
    """
    Toggle the bit in the given value.

    Args:
        value (int): the integer value that needs a bit toggled
        bit (int): the index of the bit to toggle, starting from 0 at the LSB

    Returns:
        the changed value.
    """
    return value ^ (1 << bit)


def bit_set(value: int, bit: int) -> bool:
    """
    Return True if the bit is set.

    Args:
        value (int): the value to check
        bit (int): the index of the bit to check, starting from 0 at the LSB

    Returns:
        True if the bit is set (1).
    """
    bit_value = 1 << bit
    return value & bit_value == bit_value


def bits_set(value: int, *args: int) -> bool:
    """
    Return True if all the bits are set.

    Args:
        value (int): the value to check
        args: a set of indices of the bits to check, starting from 0 at the LSB.
            All the indices shall be given as separate arguments, i.e. unpack a list
            if needed.

    Returns:
        True if all the bits are set (1).

    Examples:
        >>> assert bits_set(0b0101_0000_1011, 0, 1, 3, 8, 10)
        >>> assert bits_set(0b0101_0000_1011, 3, 8)
        >>> assert not bits_set(0b0101_0000_1011, *[1, 2, 3])
    """

    return all([bit_set(value, bit) for bit in args])


def beautify_binary(value: int, sep: str = " ", group: int = 8, prefix: str = "", size: int = 0) -> str:
    """
    Returns a binary representation of the given value. The bits are presented
    in groups of 8 bits for clarity by default (can be changed with the `group` keyword).

    Args:
        value (int): the value to beautify
        sep (str): the separator character to be used, default is a space
        group (int): the number of bits to group together, default is 8
        prefix (str): a string to prefix the result, default is ''
        size (int): number of digits

    Returns:
        a binary string representation.

    Examples:
        >>> status = 2**14 + 2**7
        >>> assert beautify_binary(status) == "01000000 10000000"
    """

    if size == 0:
        size = 8
        while value > 2**size - 1:
            size += 8

    b_str = f"{value:0{size}b}"

    return prefix + sep.join([b_str[i : i + group] for i in range(0, len(b_str), group)])


def humanize_bytes(n: int, base: Union[int, str] = 2, precision: int = 3) -> str:
    """
    Represents the size `n` in human-readable form, i.e. as byte, KiB, MiB, GiB, ...

    Args:
        n (int): number of byte
        base (int, str): binary (2) or decimal (10)
        precision (int): the number of decimal places [default=3]

    Returns:
        a human-readable size, like 512 byte or 2.300 TiB

    Raises:
        ValueError: when base is different from 2 (binary) or 10 (decimal).

    Examples:
        >>> assert humanize_bytes(55) == "55 bytes"
        >>> assert humanize_bytes(1024) == "1.000 KiB"
        >>> assert humanize_bytes(1000, base=10) == "1.000 kB"
        >>> assert humanize_bytes(1000000000) == '953.674 MiB'
        >>> assert humanize_bytes(1000000000, base=10) == '1.000 GB'
        >>> assert humanize_bytes(1073741824) == '1.000 GiB'
        >>> assert humanize_bytes(1024**5 - 1, precision=0) == '1024 TiB'

    Note:
        Please note that, by default, I use the IEC standard (International Engineering Consortium)
        which is in `base=2` (binary), i.e. 1024 bytes = 1.0 KiB. If you need SI units (International
        System of Units), you need to specify `base=10` (decimal), i.e. 1000 bytes = 1.0 kB.

    """

    if base not in [2, 10, "binary", "decimal"]:
        raise ValueError(f"Only base 2 (binary) and 10 (decimal) are supported, got {base}.")

    # By default, we assume base == 2 or base == "binary"

    one_kilo = 1024
    units = ["KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]

    if base == 10 or base == "decimal":
        one_kilo = 1000
        units = ["kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]

    _n: float = n
    if _n < one_kilo:
        return f"{_n} byte{'' if n == 1 else 's'}"

    for dim in units:
        _n /= one_kilo
        if _n < one_kilo:
            return f"{_n:.{precision}f} {dim}"

    return f"{n} byte{'' if n == 1 else 's'}"


CRC_TABLE = [
    0x00,
    0x91,
    0xE3,
    0x72,
    0x07,
    0x96,
    0xE4,
    0x75,
    0x0E,
    0x9F,
    0xED,
    0x7C,
    0x09,
    0x98,
    0xEA,
    0x7B,
    0x1C,
    0x8D,
    0xFF,
    0x6E,
    0x1B,
    0x8A,
    0xF8,
    0x69,
    0x12,
    0x83,
    0xF1,
    0x60,
    0x15,
    0x84,
    0xF6,
    0x67,
    0x38,
    0xA9,
    0xDB,
    0x4A,
    0x3F,
    0xAE,
    0xDC,
    0x4D,
    0x36,
    0xA7,
    0xD5,
    0x44,
    0x31,
    0xA0,
    0xD2,
    0x43,
    0x24,
    0xB5,
    0xC7,
    0x56,
    0x23,
    0xB2,
    0xC0,
    0x51,
    0x2A,
    0xBB,
    0xC9,
    0x58,
    0x2D,
    0xBC,
    0xCE,
    0x5F,
    0x70,
    0xE1,
    0x93,
    0x02,
    0x77,
    0xE6,
    0x94,
    0x05,
    0x7E,
    0xEF,
    0x9D,
    0x0C,
    0x79,
    0xE8,
    0x9A,
    0x0B,
    0x6C,
    0xFD,
    0x8F,
    0x1E,
    0x6B,
    0xFA,
    0x88,
    0x19,
    0x62,
    0xF3,
    0x81,
    0x10,
    0x65,
    0xF4,
    0x86,
    0x17,
    0x48,
    0xD9,
    0xAB,
    0x3A,
    0x4F,
    0xDE,
    0xAC,
    0x3D,
    0x46,
    0xD7,
    0xA5,
    0x34,
    0x41,
    0xD0,
    0xA2,
    0x33,
    0x54,
    0xC5,
    0xB7,
    0x26,
    0x53,
    0xC2,
    0xB0,
    0x21,
    0x5A,
    0xCB,
    0xB9,
    0x28,
    0x5D,
    0xCC,
    0xBE,
    0x2F,
    0xE0,
    0x71,
    0x03,
    0x92,
    0xE7,
    0x76,
    0x04,
    0x95,
    0xEE,
    0x7F,
    0x0D,
    0x9C,
    0xE9,
    0x78,
    0x0A,
    0x9B,
    0xFC,
    0x6D,
    0x1F,
    0x8E,
    0xFB,
    0x6A,
    0x18,
    0x89,
    0xF2,
    0x63,
    0x11,
    0x80,
    0xF5,
    0x64,
    0x16,
    0x87,
    0xD8,
    0x49,
    0x3B,
    0xAA,
    0xDF,
    0x4E,
    0x3C,
    0xAD,
    0xD6,
    0x47,
    0x35,
    0xA4,
    0xD1,
    0x40,
    0x32,
    0xA3,
    0xC4,
    0x55,
    0x27,
    0xB6,
    0xC3,
    0x52,
    0x20,
    0xB1,
    0xCA,
    0x5B,
    0x29,
    0xB8,
    0xCD,
    0x5C,
    0x2E,
    0xBF,
    0x90,
    0x01,
    0x73,
    0xE2,
    0x97,
    0x06,
    0x74,
    0xE5,
    0x9E,
    0x0F,
    0x7D,
    0xEC,
    0x99,
    0x08,
    0x7A,
    0xEB,
    0x8C,
    0x1D,
    0x6F,
    0xFE,
    0x8B,
    0x1A,
    0x68,
    0xF9,
    0x82,
    0x13,
    0x61,
    0xF0,
    0x85,
    0x14,
    0x66,
    0xF7,
    0xA8,
    0x39,
    0x4B,
    0xDA,
    0xAF,
    0x3E,
    0x4C,
    0xDD,
    0xA6,
    0x37,
    0x45,
    0xD4,
    0xA1,
    0x30,
    0x42,
    0xD3,
    0xB4,
    0x25,
    0x57,
    0xC6,
    0xB3,
    0x22,
    0x50,
    0xC1,
    0xBA,
    0x2B,
    0x59,
    0xC8,
    0xBD,
    0x2C,
    0x5E,
    0xCF,
]


def crc_calc(data: list[bytes | int], start: int, len_: int) -> int:
    """
    Calculate the checksum for (part of) the data.

    Args:
        data: the data for which the checksum needs to be calculated
        start: offset into the data array [byte]
        len_: number of bytes to incorporate into the calculation

    Returns:
        the calculated checksum.

    Reference:
        The description of the CRC calculation for RMAP is given in the ECSS document
        _Space Engineering: SpaceWire - Remote Memory Access Protocol_, section A.3
        on page 80 [ECSS-E-ST-50-52C].

    """
    crc: int = 0

    # The check below is needed because we can pass data that is of type ctypes.c_char_Array
    # and the individual elements have then type 'bytes'.

    if isinstance(data[0], bytes):
        for idx in range(start, start + len_):
            crc = CRC_TABLE[crc ^ (int.from_bytes(data[idx], byteorder="big") & 0xFF)]
    elif isinstance(data[0], int):
        for idx in range(start, start + len_):
            crc = CRC_TABLE[crc ^ (data[idx] & 0xFF)]
    else:
        ...

    return crc


def s16(value: int) -> int:
    """
    Return the signed equivalent of a hex or binary number.

    Args:
        value: an integer value.

    Returns:
        The negative equivalent of a twos-complement binary number.

    Examples:
        Since integers in Python are objects and stored in a variable number of bits, Python doesn't
        know the concept of twos-complement for negative integers. For example, this 16-bit number

        >>> 0b1000_0000_0001_0001
        32785

        which in twos-complement is actually a negative value:

        >>> s16(0b1000_0000_0001_0001)
        -32751

        The 'bin()' function will return a strange representation of this number:

        >>> bin(-32751)
        '-0b111111111101111'

        when we however mask the value we get:

        >>> bin(-32751 & 0b1111_1111_1111_1111)
        '0b1000000000010001'

    See:
        [Twos complement in Python](https://stackoverflow.com/questions/1604464/twos-complement-in-python) and \
        [Pythons representation of negative integers](
        https://stackoverflow.com/questions/46993519/python-representation-of-negative-integers) and \
        [Signed equivalent of a twos-complement hex-value](
        https://stackoverflow.com/questions/25096755/signed-equivalent-of-a-2s-complement-hex-value) and \
        [SO Twos complement in Python](https://stackoverflow.com/a/32262478/4609203).

    """
    return ctypes.c_int16(value).value


def s32(value: int) -> int:
    """
    Return the signed equivalent of a hex or binary number.

    Args:
        value: an integer value.

    Returns:
        The negative equivalent of a twos-complement binary number.

    Examples:
        Since integers in Python are objects and stored in a variable number of bits, Python doesn't
        know the concept of twos-complement for negative integers. For example, this 32-bit number

        >>> 0b1000_0000_0000_0000_0000_0000_0001_0001
        2147483665

        which in twos-complement is actually a negative value:

        >>> s32(0b1000_0000_0000_0000_0000_0000_0001_0001)
        -2147483631

    """
    return ctypes.c_int32(value).value



"""
This module defines base classes and generic functions to work with sockets.
"""

import asyncio
import select
import socket
import time
from typing import Optional

from egse.device import AsyncDeviceInterface
from egse.device import AsyncDeviceTransport
from egse.device import DeviceConnectionError
from egse.device import DeviceConnectionInterface
from egse.device import DeviceTimeoutError
from egse.device import DeviceTransport
from egse.log import logger
from egse.system import type_name

SEPARATOR = b"\x03"
SEPARATOR_STR = SEPARATOR.decode()


class SocketDevice(DeviceConnectionInterface, DeviceTransport):
    """Base class that implements the socket interface."""

    # We set a default connect timeout of 3.0 sec before connecting and reset
    # to None (=blocking) after connecting. The reason for this is that when no
    # device is available, e.g. during testing, the timeout will take about
    # two minutes which is way too long. It needs to be evaluated if this
    # approach is acceptable and not causing problems during production.

    def __init__(
        self,
        hostname: str,
        port: int,
        connect_timeout: float = 3.0,
        read_timeout: float | None = 1.0,
        separator: bytes = SEPARATOR,
    ):
        super().__init__()
        self.is_connection_open = False
        self.hostname = hostname
        self.port = port
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.separator = separator
        self.separator_str = separator.decode()
        self.socket = None

    @property
    def device_name(self):
        """The name of the device that this interface connects to."""
        return f"SocketDevice({self.hostname}:{self.port})"

    def connect(self):
        """
        Connect the device.

        Raises:
            ConnectionError: When the connection could not be established.
                Check the logging messages for more detail.
            TimeoutError: When the connection timed out.
            ValueError: When hostname or port number are not provided.
        """

        # Sanity checks

        if self.is_connection_open:
            logger.warning(f"{self.device_name}: trying to connect to an already connected socket.")
            return

        if self.hostname in (None, ""):
            raise ValueError(f"{self.device_name}: hostname is not initialized.")

        if self.port in (None, 0):
            raise ValueError(f"{self.device_name}: port number is not initialized.")

        # Create a new socket instance

        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as exc:
            raise ConnectionError(f"{self.device_name}: Failed to create socket.") from exc

        try:
            logger.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self.socket.settimeout(self.connect_timeout)
            self.socket.connect((self.hostname, self.port))
            self.socket.settimeout(None)
        except ConnectionRefusedError as exc:
            raise ConnectionError(f"{self.device_name}: Connection refused to {self.hostname}:{self.port}.") from exc
        except TimeoutError as exc:
            raise TimeoutError(f"{self.device_name}: Connection to {self.hostname}:{self.port} timed out.") from exc
        except socket.gaierror as exc:
            raise ConnectionError(f"{self.device_name}: Socket address info error for {self.hostname}") from exc
        except socket.herror as exc:
            raise ConnectionError(f"{self.device_name}: Socket host address error for {self.hostname}") from exc
        except OSError as exc:
            raise ConnectionError(f"{self.device_name}: OSError caught ({exc}).") from exc

        self.is_connection_open = True

    def disconnect(self):
        """
        Disconnect from the Ethernet connection.

        Raises:
            ConnectionError when the socket could not be closed.
        """

        assert self.socket is not None  # extra check + for mypy type checking

        try:
            if self.is_connection_open:
                logger.debug(f"Disconnecting from {self.hostname}")
                self.socket.close()
                self.is_connection_open = False
        except Exception as e_exc:
            raise ConnectionError(f"{self.device_name}: Could not close socket to {self.hostname}") from e_exc

    def is_connected(self) -> bool:
        """
        Check if the device is connected.

        Returns:
             True is the device is connected, False otherwise.
        """

        return bool(self.is_connection_open)

    def reconnect(self):
        """
        Reconnect to the device. If the connection is open, this function will first disconnect
        and then connect again.
        """

        if self.is_connection_open:
            self.disconnect()
        self.connect()

    def read(self) -> bytes:
        """
        Read until ETX (b'\x03') or until `self.read_timeout` elapses.
        Uses `select` to avoid blocking indefinitely when no data is available.
        If `self.read_timeout` was set to None in the constructor, this will block anyway.
        """
        if not self.socket:
            raise DeviceConnectionError(self.device_name, "The device is not connected, connect before reading.")

        buf_size = 1024 * 4
        response = bytearray()

        # If read_timeout is None we preserve blocking behavior; otherwise enforce overall timeout.
        if self.read_timeout is None:
            end_time = None
        else:
            end_time = time.monotonic() + self.read_timeout

        try:
            while True:
                # compute the remaining timeout for select, this is needed because we read in different parts
                # until ETX is received, and we want to receive the complete messages including ETX within
                # the read timeout.
                if end_time is None:
                    timeout = None
                else:
                    remaining = end_time - time.monotonic()
                    if remaining <= 0.0:
                        raise DeviceTimeoutError(self.device_name, "Socket read timed out")
                    timeout = remaining

                ready, _, _ = select.select([self.socket], [], [], timeout)

                if not ready:
                    # no socket ready within timeout
                    raise DeviceTimeoutError(self.device_name, "Socket read timed out")

                try:
                    data = self.socket.recv(buf_size)
                except OSError as exc:
                    raise DeviceConnectionError(self.device_name, f"Caught {type_name(exc)}: {exc}") from exc

                if not data:
                    # remote closed connection (EOF)
                    raise DeviceConnectionError(self.device_name, "Connection closed by peer")

                response.extend(data)

                if self.separator in response:
                    break

        except DeviceTimeoutError:
            raise
        except DeviceConnectionError:
            raise
        except Exception as exc:
            # unexpected errors - translate to DeviceConnectionError
            raise DeviceConnectionError(self.device_name, "Socket read error") from exc

        return bytes(response)

    def write(self, command: str):
        """
        Send a command to the device.

        No processing is done on the command string, except for the encoding into a bytes object.

        Args:
            command: the command string including terminators.

        Raises:
            A DeviceTimeoutError when the send timed out, and a DeviceConnectionError if
            there was a socket related error.
        """

        if not self.socket:
            raise DeviceConnectionError(self.device_name, "The device is not connected, connect before writing.")

        try:
            command += self.separator_str if not command.endswith(self.separator_str) else ""
            self.socket.sendall(command.encode())
        except socket.timeout as exc:
            raise DeviceTimeoutError(self.device_name, "Socket timeout error") from exc
        except socket.error as exc:
            # Interpret any socket-related error as an I/O error
            raise DeviceConnectionError(self.device_name, "Socket communication error.") from exc

    def trans(self, command: str) -> bytes:
        """
        Send a command to the device and wait for the response.

        No processing is done on the command string, except for the encoding into a bytes object.

        Args:
            command: the command string including terminators.

        Returns:
            A bytes object containing the response from the device. No processing is done
            on the response.

        Raises:
            A DeviceTimeoutError when the send timed out, and a DeviceConnectionError if
            there was a socket related error.
        """

        if not self.socket:
            raise DeviceConnectionError(self.device_name, "The device is not connected, connect before writing.")

        self.write(command)

        response = self.read()

        return response


class AsyncSocketDevice(AsyncDeviceInterface, AsyncDeviceTransport):
    """
    Async socket-backed device using asyncio streams.

    - async connect() / disconnect()
    - async read() reads until ETX (b'\\x03') or timeout
    - async write() and async trans()
    """

    def __init__(
        self,
        hostname: str,
        port: int,
        connect_timeout: float = 3.0,
        read_timeout: float | None = 1.0,
        separator: bytes = SEPARATOR,
    ):
        super().__init__()
        self.hostname = hostname
        self.port = port
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout
        self.separator = separator
        self.reader: Optional[asyncio.StreamReader] = None
        self.writer: Optional[asyncio.StreamWriter] = None
        self.is_connection_open = False

    @property
    def device_name(self) -> str:
        # Override this property for a decent name
        return f"AsyncSocketDevice({self.hostname}:{self.port})"

    async def connect(self) -> None:
        if self.is_connection_open:
            logger.debug(f"{self.device_name}: already connected")
            return

        if not self.hostname:
            raise ValueError(f"{self.device_name}: hostname is not initialized.")
        if not self.port:
            raise ValueError(f"{self.device_name}: port is not initialized.")

        try:
            logger.debug(f"{self.device_name}: connect() called; is_connection_open={self.is_connection_open}")
            coro = asyncio.open_connection(self.hostname, self.port)
            self.reader, self.writer = await asyncio.wait_for(coro, timeout=self.connect_timeout)
            self.is_connection_open = True
            logger.debug(f"{self.device_name}: connected -> peer={self.writer.get_extra_info('peername')}")

        except asyncio.TimeoutError as exc:
            await self._cleanup()
            logger.warning(f"{self.device_name}: connect timed out")
            raise DeviceTimeoutError(self.device_name, f"Connection to {self.hostname}:{self.port} timed out.") from exc
        except Exception as exc:
            await self._cleanup()
            logger.warning(f"{self.device_name}: connect failed: {type_name(exc)} – {exc}")
            raise DeviceConnectionError(self.device_name, f"Failed to connect to {self.hostname}:{self.port}") from exc

    async def disconnect(self) -> None:
        logger.debug(f"{self.device_name}: disconnect() called; writer_exists={bool(self.writer)}")
        peer = None
        try:
            if self.writer and not self.writer.is_closing():
                peer = self.writer.get_extra_info("peername")
                self.writer.close()
                # wait for close, but don't hang forever
                try:
                    await asyncio.wait_for(self.writer.wait_closed(), timeout=1.0)
                except asyncio.TimeoutError:
                    logger.debug(f"{self.device_name}: wait_closed() timed out for peer={peer}")

        finally:
            await self._cleanup()
            logger.debug(f"{self.device_name}: disconnected ({peer=})")

    def is_connected(self) -> bool:
        return bool(self.is_connection_open and self.writer and not self.writer.is_closing())

    async def _cleanup(self) -> None:
        self.reader = None
        self.writer = None
        self.is_connection_open = False

    async def read(self) -> bytes:
        if not self.reader:
            raise DeviceConnectionError(self.device_name, "Not connected")
        try:
            # readuntil includes the separator; we keep it for parity with existing code
            data = await asyncio.wait_for(self.reader.readuntil(separator=self.separator), timeout=self.read_timeout)
            return data
        except asyncio.IncompleteReadError as exc:
            # EOF before separator
            await self._cleanup()
            raise DeviceConnectionError(self.device_name, "Connection closed while reading") from exc
        except asyncio.TimeoutError as exc:
            raise DeviceTimeoutError(self.device_name, "Socket read timed out") from exc
        except Exception as exc:
            await self._cleanup()
            raise DeviceConnectionError(self.device_name, "Socket read error") from exc

    async def write(self, command: str) -> None:
        if not self.writer:
            raise DeviceConnectionError(self.device_name, "Not connected")
        try:
            self.writer.write(command.encode())
            await asyncio.wait_for(self.writer.drain(), timeout=self.read_timeout)
        except asyncio.TimeoutError as exc:
            raise DeviceTimeoutError(self.device_name, "Socket write timed out") from exc
        except Exception as exc:
            await self._cleanup()
            raise DeviceConnectionError(self.device_name, "Socket write error") from exc

    async def trans(self, command: str) -> bytes:
        if not self.writer or not self.reader:
            raise DeviceConnectionError(self.device_name, "Not connected")
        try:
            await self.write(command)
            return await self.read()
        except (DeviceTimeoutError, DeviceConnectionError):
            raise
        except Exception as exc:
            await self._cleanup()
            raise DeviceConnectionError(self.device_name, "Socket trans error") from exc



from __future__ import annotations

__all__ = [
    "HKError",
    "TmDictionaryColumns",
    "convert_hk_names",
    "get_housekeeping",
    "get_hk_info",
    "read_conversion_dict",
]
import csv
import datetime
from enum import Enum
from pathlib import Path
from typing import Optional
from typing import Union

import dateutil.parser as date_parser
import numpy as np
import pandas as pd

from egse.config import find_files
from egse.env import get_data_storage_location
from egse.env import get_site_id
from egse.log import logger
from egse.obsid import ObservationIdentifier
from egse.obsid import obsid_from_storage
from egse.setup import Setup
from egse.setup import SetupError
from egse.setup import load_setup
from egse.system import SECONDS_IN_A_DAY
from egse.system import read_last_line
from egse.system import read_last_lines
from egse.system import str_to_datetime
from egse.system import time_since_epoch_1958


class TmDictionaryColumns(str, Enum):
    """Enumeration of the relevant columns in the TM dictionary spreadsheet.

    The relevant columns are:

    | Name | Description |
    | ---- | ----------- |
    | STORAGE_MNEMONIC | Column with the storage mnemonic of the process that generated the HK |
    | CORRECT_HK_NAMES | Column with the correct HK names (that can be used in `get_housekeeping`) |
    | ORIGINAL_EGSE_HK_NAMES | Column with the names that were originally used in `get_housekeeping` the device protocol |
    | SYNOPTICS_ORIGIN | Column with the origin of the synoptics at the current site |
    | TIMESTAMP_NAMES | Column with the name of the timestamps |
    | DASHBOARD | Column with the name of the dashboard that holds the HK metric |
    """  # noqa

    STORAGE_MNEMONIC = "Storage mnemonic"
    CORRECT_HK_NAMES = "CGSE mnemonic"
    ORIGINAL_EGSE_HK_NAMES = "Original name in CGSE"
    SYNOPTICS_ORIGIN = f"Origin of synoptics at {get_site_id()}"
    TIMESTAMP_NAMES = "Name of corresponding timestamp"
    DESCRIPTION = "Description"
    DASHBOARD = "MON screen"


class HKError(Exception):
    """An HK-specific error."""

    pass


def get_housekeeping(
    hk_name: str,
    obsid: Union[ObservationIdentifier, str, int] = None,
    od: str = None,
    time_window: int = None,
    data_dir: str = None,
    setup: Optional[Setup] = None,
) -> tuple | np.ndarray:
    """
    Returns the timestamp(s) and housekeeping value(s) for the housekeeping parameter with the given name.

    It is possible to indicate for which obsid or which OD the housekeeping is to be returned.  If neither of them is
    specified, the latest daily files are used.

    When the time window has not been specified, the last timestamp and housekeeping value will be returned for the
    given OD. It is possible that a component stopped writing HK for some reason, and that the last housekeeping value
    is older than you would want.  It is therefore important to inspect the corresponding timestamp.

    When the time window has been specified, the relevant housekeeping will be read:

    * determine the sampling rate (compare the timestamps for the last 2 lines in the housekeeping file);
    * determine how many samples we need to read (starting at the back);
    * read the required number of line, starting at the back;
    * for each of the read lines, append the timestamp and HK value to the arrays that will be returned

    Args:
        hk_name: Name of the housekeeping parameter.
        obsid: Observation identifier.  This can be an ObservationIdentifier object, a string in format TEST_LAB or
            TEST_LAB_SETUP, or an integer representing the test ID; optional.
        od: Identifier for the OD (YYYYMMDD); optional.
        time_window: Length of the time window over which to retrieve the housekeeping [s].  The time window ends at
            the moment this method is called.  If not given, the latest housekeeping value is returned.
        data_dir: Folder (with sub-folders /daily and /obs) in which the HK files are stored. If this argument is not
            provided, the data_dir will be determined from the environment variable `${PROJECT}_DATA_STORAGE_LOCATION`.
        setup: the Setup to use, if None, the setup will be loaded from the configuration manager

    Raises:
        HKError: when one of the following problems occur
            * no obsid nor an od argument was provided
            * no HK measures were found for the given parameter and obsid/od

    Returns:
        A tuple or an array based on the time window:
            - If the time window has not been specified: the most recent timestamp and housekeeping value.
            - If the time window has been specified: an array of timestamps and an array of housekeeping values,
              belonging to the specified time window.

    """

    setup = setup or load_setup()

    # Either specify the obsid or the OD (or neither of them) but not both

    if obsid is not None and od is not None:
        raise HKError(f"Both the obsid ({obsid}) and the OD ({od}) were specified.")

    # Specified obsid (as integer or as string)

    data_dir = data_dir or get_data_storage_location()

    if obsid:
        try:
            return _get_housekeeping_obsid(hk_name, data_dir, obsid=obsid, time_window=time_window, setup=setup)
        except (ValueError, StopIteration, FileNotFoundError) as exc:
            raise HKError(f"No HK found for {hk_name} for obsid {obsid} at {get_site_id()}") from exc

    # Specified OD

    if od:
        try:
            return _get_housekeeping_od(hk_name, data_dir, od=od, time_window=time_window, setup=setup)
        except (ValueError, StopIteration, FileNotFoundError) as exc:
            raise HKError(f"No HK found for {hk_name} for OD {od} at {get_site_id()}") from exc

    # Didn't specify neither the obsid nor the OD

    try:
        return _get_housekeeping_daily(hk_name, data_dir, time_window=time_window, setup=setup)
    except (ValueError, StopIteration, FileNotFoundError) as exc:
        raise HKError(f"No HK found for {hk_name} for today at {get_site_id()}") from exc


def _get_housekeeping(hk_name: str, timestamp_name: str, hk_dir: str, files, time_window: int = None):
    """Return the timestamp(s) and HK value(s) for the HK parameter with the given name, for the given files.

    When the time window has not been specified, the last timestamp and HK value will be returned for the given OD.
    It is possible that a component stopped writing HK for some reason, and that the last HK value is older than you
    would want.  It is therefore important to inspect the corresponding timestamp.

    When the time window has been specified, the relevant HK will be read:
        - determine the sampling rate (compare the timestamps for the last 2 lines in the HK file);
        - determine how many samples we need to read (starting at the back);
        - read the required number of line, starting at the back;
        - for each of the read lines, append the timestamp and HK value to the arrays that will be returned

    Args:
        - hk_name: Name of the housekeeping parameter.
        - timestamp_name: Name of the corresponding timestamp.
        - hk_dir: Directory with the housekeeping files.
        - files: Relative filepath of the selected housekeeping files.
        - time_window: Length of the time window over which to retrieve the housekeeping [s].  The time window ends at
                       the moment this method is called.  If not given, the latest HK-value is returned.

    Returns:
        - If the time window has not been specified: the most recent timestamp and housekeeping value.
        - If the time window has been specified: an array of timestamps and an array of housekeeping values, belonging
          to the specified time window.
    """

    filename = files[-1]

    # Indices of the columns we need

    timestamp_index, hk_index = get_indices(hk_dir + filename, hk_name, timestamp_name)

    # No time window specified: return the last value

    if time_window is None:
        return get_last_non_empty(hk_dir + filename, timestamp_index, hk_index)

    # Time window specified

    else:
        # We will return an array of timestamps and an array of HK values

        timestamp_array = np.array([])
        hk_array = np.array([])

        with open(hk_dir + filename) as file:
            csv_reader = csv.reader(file)
            next(csv_reader)
            first_timepoint = next(csv_reader)[0].split(",")[timestamp_index]  # Skip the header

        last_timepoint = read_last_line(hk_dir + filename).split(",")[timestamp_index]
        elapsed = (str_to_datetime(last_timepoint) - str_to_datetime(first_timepoint)).total_seconds()

        # The time window is shorter than the timespan covered by the file

        if time_window < elapsed:
            sampling_rate = get_sampling_rate(hk_dir + filename, timestamp_name)  # Time between subsequent samples
            num_samples = int(round(time_window / sampling_rate))

            lines = read_last_lines(hk_dir + filename, num_samples)

            for line in lines:
                line = line.split(",")

                timestamp_array = np.append(timestamp_array, line[timestamp_index])
                hk_array = np.append(hk_array, line[hk_index])

        # The time window is longer than the timespan covered by the file: read all lines

        else:
            with open(hk_dir + filename) as file:
                csv_reader = csv.reader(file)
                next(csv_reader)  # Skip the header

                for row in csv_reader:
                    timestamp_array = np.append(timestamp_array, row[timestamp_index])
                    hk_array = np.append(hk_array, row[hk_index])

        for index in range(len(timestamp_array)):
            timestamp_array[index] = time_since_epoch_1958(timestamp_array[index])

        return timestamp_array, hk_array


def _get_housekeeping_od(hk_name: str, data_dir, od: str, time_window: int = None, setup: Optional[Setup] = None):
    """Return the timestamp(s) and HK value(s) for the HK parameter with the given name, for the given OD.

    When the time window has not been specified, the last timestamp and HK value will be returned for the given OD.
    It is possible that a component stopped writing HK for some reason, and that the last HK value is older than you
    would want.  It is therefore important to inspect the corresponding timestamp.

    When the time window has been specified, the relevant HK will be read:
        - determine the sampling rate (compare the timestamps for the last 2 lines in the HK file);
        - determine how many samples we need to read (starting at the back);
        - read the required number of line, starting at the back;
        - for each of the read lines, append the timestamp and HK value to the arrays that will be returned

    Args:
        - hk_name: Name of the housekeeping parameter.
        - data_dir: Folder (with sub-folders /daily and /obs) in which the HK files are stored.
        - od: Identifier for the OD (YYYYMMDD).
        - time_window: Length of the time window over which to retrieve the housekeeping [s].  The time window ends at
                       the moment this method is called.  If not given, the latest HK-value is returned.
        - setup: Setup.

        Returns:
            - If the time window has not been specified: the most recent timestamp and HK value.
            - If the time window has been specified: an array of timestamps and an array of HK values, belonging to the
              specified time window.
    """

    setup = setup or load_setup()
    hk_dir = f"{data_dir}/daily/"  # Where the HK is stored

    try:
        origin, timestamp_name = get_hk_info(hk_name, setup=setup)

    except KeyError:
        raise HKError(f"Cannot determine which EGSE component generated HK parameter {hk_name}")

    hk_dir += f"{od}/"
    hk_files = [f"{od}_{get_site_id()}_{origin}.csv"]

    return _get_housekeeping(hk_name, timestamp_name, hk_dir, hk_files, time_window=time_window)


def _get_housekeeping_obsid(
    hk_name: str,
    data_dir,
    obsid: Union[ObservationIdentifier, str, int],
    time_window: int = None,
    setup: Optional[Setup] = None,
):
    """Return the timestamp(s) and HK value(s) for the HK parameter with the given name, for the given obsid.

    When the time window has not been specified, the last timestamp and HK value will be returned for the given obsid.
    It is possible that a component stopped writing HK for some reason, and that the last HK value is older than you
    would want.  It is therefore important to inspect the corresponding timestamp.

    When the time window has been specified, the relevant HK will be read:
        - determine the sampling rate (compare the timestamps for the last 2 lines in the HK file);
        - determine how many samples we need to read (starting at the back);
        - read the required number of line, starting at the back;
        - for each of the read lines, append the timestamp and HK value to the arrays that will be returned

    Args:
        - hk_name: Name of the housekeeping parameter.
        - data_dir: Folder (with sub-folders /daily and /obs) in which the HK files are stored.
        - obsid: Observation identifier.  This can be an ObservationIdentifier object, a string in format TEST_LAB or
                 TEST_LAB_SETUP, or an integer representing the test ID.
        - time_window: Length of the time window over which to retrieve the housekeeping [s].  The time window ends at
                       the moment this method is called.  If not given, the latest HK-value is returned.
        - setup: Setup.

    Returns:
        - If the time window has not been specified: the most recent timestamp and HK value.
        - If the time window has been specified: an array of timestamps and an array of HK values, belonging to the
          specified time window.
    """

    setup = setup or load_setup()

    hk_dir = f"{data_dir}/obs/"  # Where the HK is stored

    try:
        origin, timestamp_name = get_hk_info(hk_name, setup=setup)

    except KeyError:
        raise HKError(f"Cannot determine which EGSE component generated HK parameter {hk_name}")

    obsid = obsid_from_storage(obsid, data_dir=data_dir)  # Convert the obsid to the correct format

    hk_dir += f"{obsid}/"
    pattern = f"{obsid}_{origin}_*.csv"
    hk_files = sorted(find_files(pattern=pattern, root=hk_dir))

    if len(hk_files) == 0:
        raise HKError(f"No HK found for the {origin} at {get_site_id()} for obsid {obsid}")

    hk_files = [hk_files[-1].name]

    return _get_housekeeping(hk_name, timestamp_name, hk_dir, hk_files, time_window=time_window)


def _get_housekeeping_daily(hk_name: str, data_dir, time_window: int = None, setup: Optional[Setup] = None):
    """Return the timestamp(s) and HK value(s) for the HK parameter with the given name.

    When the time window has not been specified, the last timestamp and HK value will be returned.  It is possible that
    a component stopped writing HK for some reason, and that the last HK value is older than you would want.  It is
    therefore important to inspect the corresponding timestamp.

    When the time window has been specified, it is possible that we have to fetch the HK from multiple files,
    depending on the length of the time window:

        * Oldest file (i.e. the HK file in which the first timestamp in the specified time window is held): only part
          of it will have to be read (start reading at the back):
            - determine the filename;
            - determine the sampling rate (compare the timestamps for the last 2 lines in the HK file);
            - determine how many samples we need to read (starting at the back);
            - read the required number of line, starting at the back;
            - for each of the read lines, append the timestamp and HK value to the arrays that will be returned.
        * Any other files: entire file will have to be read:
            - determine the filename;
            - read all lines;
            - for each line: append the timestamp and HK value to the arrays that will be returned.

    Args:
        - hk_name: Name of the housekeeping parameter.
        - data_dir: Folder (with sub-folders /daily and /obs) in which the HK files are stored.
        - time_window: Length of the time window over which to retrieve the housekeeping [s].  The time window ends at
          the moment this method is called.  If not given, the latest HK-value is returned.
        - setup: Setup.

    Returns:
        - If the time window has not been specified: the most recent timestamp and HK value.
        - If the time window has been specified: an array of timestamps and an array of HK values, belonging to the
          specified time window.
    """

    setup = setup or load_setup()
    hk_dir = f"{data_dir}/daily/"  # Where the HK is stored

    try:
        origin, timestamp_name = get_hk_info(hk_name, setup=setup)

    except KeyError:
        raise HKError(f"Cannot determine which EGSE component generated HK parameter {hk_name}")

    # No time window specified: return the last value

    if time_window is None:
        # Look for the last file of this component

        timestamp = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y%m%d")
        hk_dir += f"{timestamp}/"
        filename = f"{timestamp}_{get_site_id()}_{origin}.csv"

        timestamp_index, hk_index = get_indices(hk_dir + filename, hk_name, timestamp_name)
        return get_last_non_empty(hk_dir + filename, timestamp_index, hk_index)

    # Time window specified

    else:
        # We will return an array of timestamps and an array of HK values

        timestamp_array = np.array([])
        hk_array = np.array([])

        # Go back in time from this very moment and determine:
        #   - which timespan the most recent HK file covers (i.e. how much time has elapsed since midnight)
        #   - what the time is at the start of the time window

        now = datetime.datetime.utcnow()
        elapsed_since_midnight = now.microsecond * 1e-6 + now.second + 60 * (now.minute + 60 * now.hour)
        start_time = now - datetime.timedelta(seconds=time_window)
        start_od = f"{start_time.year}{start_time.month:02d}{start_time.day:02d}"

        # Determine which columns will be needed from which file

        filename = f"{start_od}/{start_od}_{get_site_id()}_{origin}.csv"

        if Path(hk_dir + filename).exists():
            timestamp_index, hk_index = get_indices(hk_dir + filename, hk_name, timestamp_name)

            # Determine how many time samples you need to read in the first relevant HK file (starting from the back)

            sampling_rate = get_sampling_rate(hk_dir + filename, timestamp_name)  # Time between subsequent samples

            if time_window <= elapsed_since_midnight:
                num_samples_first_day = int(round(time_window / sampling_rate))

            else:
                time_window_first_day = (time_window - elapsed_since_midnight) % SECONDS_IN_A_DAY
                num_samples_first_day = int(round(time_window_first_day / sampling_rate))  # TODO Round or floor?

            # Read the required number of lines in the relevant HK file (starting from the back of the file)

            lines_first_day = read_last_lines(hk_dir + filename, num_samples_first_day)

            for line in lines_first_day:
                line = line.split(",")

                timestamp_array = np.append(timestamp_array, line[timestamp_index])
                hk_array = np.append(hk_array, line[hk_index])

        # In case we also need to read more recent files
        # (those will have to be read entirely)

        else:
            logger.warning(f"No HK available for {origin} on {start_time.day}/{start_time.month}/{start_time.year}")

        day = (start_time + datetime.timedelta(days=1)).date()  # The day after the first day
        last_day = datetime.date(now.year, now.month, now.day)  # Today

        while day <= last_day:
            od = f"{day.year}{day.month:02d}{day.day:02d}"
            filename = f"{od}/{od}_{get_site_id()}_{origin}.csv"

            if Path(hk_dir + filename).exists():
                with open(hk_dir + filename) as file:
                    csv_reader = csv.reader(file)

                    header = next(csv_reader)  # Skip the header
                    timestamp_index = header.index("timestamp")
                    try:
                        hk_index = header.index(hk_name)
                    except ValueError:
                        raise HKError(f"Cannot find column {hk_name} in {filename}")

                    for row in csv_reader:
                        timestamp_array = np.append(timestamp_array, row[timestamp_index])
                        hk_array = np.append(hk_array, row[hk_index])

            else:
                logger.warning(f"No HK available for {origin} on {day.day}/{day.month}/{day.year}")

            day += datetime.timedelta(days=1)

        for index in range(len(timestamp_array)):
            timestamp_array[index] = time_since_epoch_1958(timestamp_array[index])

        return timestamp_array, hk_array


def get_last_non_empty(filename: str, timestamp_index: int, hk_index: int) -> tuple:
    """Return the timestamp and HK value for last real value.

    Args:
         filename: HK file in which to look for the given HK parameter.
         timestamp_index: Index of the column with the timestamps.
         hk_index: Index of the column with the HK parameter with the given name.

    Returns:
        The timestamp and HK value with the last real value.
    """

    timestamp = None
    hk_value = " "

    filename = Path(filename)

    if not filename.exists():
        return None

    # Declaring variable to implement exponential search

    try:
        num_lines = 1

        while hk_value == " " or hk_value == "":
            pos = num_lines + 1

            # List to store last N lines

            lines = []

            with open(filename) as f:
                while len(lines) <= num_lines:
                    try:
                        f.seek(-pos, 2)

                    except IOError:
                        f.seek(0)
                        break

                    finally:
                        lines = list(f)

                    # Increasing value of variable exponentially

                    pos *= 2

            last_line = lines[-num_lines].rstrip("\r").split(",")
            timestamp, hk_value = last_line[timestamp_index], last_line[hk_index]

            num_lines += 1

        return time_since_epoch_1958(timestamp), hk_value

    except IndexError:
        return None, None


def get_indices(filename: str, hk_name: str, timestamp_name: str) -> tuple:
    """Return the column number of the timestamp and given HK parameter in the given HK file.

    Args:
        filename: HK file in which to look for the given HK parameter.
        hk_name: Name of the HK parameter.
        timestamp_name: Name of the corresponding timestamp.

    Returns:
        A tuple with:

            - Index of the column with the timestamps.
            - Index of the column with the HK parameter with the given name.
    """

    with open(filename, "r") as f:
        reader = csv.reader(f)
        header = next(reader)  # Skip the header

    timestamp_index = header.index(timestamp_name)
    # timestamp_index = 0

    try:
        hk_index = header.index(hk_name)

    except ValueError:
        raise HKError(f"Cannot find column {hk_name} in {filename}")

    return timestamp_index, hk_index


def get_sampling_rate(filename: str, timestamp_name: str) -> float:
    """Return the sampling rate for the HK file with the given name [s].

    The sampling rate is determined as the difference between the timestamps of the last two lines of the HK file.

    Args:
        filename: Name of the HK file.  We do not check explicitly whether this file exists.
        timestamp_name: the name of the column containing the timestamp

    Returns:
        Sampling rate for the HK file with the given name [s].
    """

    # Determine which column comprises the timestamp

    with open(filename, "r") as f:
        reader = csv.reader(f)
        header = next(reader)  # Skip the header

    timestamp_index = header.index(timestamp_name)

    # Read the last 2 lines and extract the timestamps for these lines

    eof = read_last_lines(filename, 2)

    penultimate_timestamp = date_parser.parse(eof[0].split(",")[timestamp_index])
    last_timestamp = date_parser.parse(eof[1].split(",")[timestamp_index])

    # Calculate the sampling rate [s]

    return (last_timestamp - penultimate_timestamp).total_seconds()


def convert_hk_names(original_hk: dict, conversion_dict: dict) -> dict:
    """
    Converts the names of the HK parameters in the given dictionary.

    The names/keys in the given dictionary of HK parameters (original_hk) are replaced by the names
    from the given conversion dictionary. The original dictionary is left unchanged, a new dictionary is
    returned.

    Args:
        original_hk: Original dictionary of HK parameters.
        conversion_dict: Dictionary with the original HK names as keys and the new HK names as values.

    Returns:
        A new dictionary of HK parameters with the corrected HK names.
    """

    converted_hk = {}

    for orig_key in original_hk:
        try:
            new_key = conversion_dict[orig_key]
        except KeyError:
            new_key = orig_key  # no conversion, just copy the key:value pair

        converted_hk[new_key] = original_hk[orig_key]

    return converted_hk


def read_conversion_dict(storage_mnemonic: str, use_site: bool = False, setup: Optional[Setup] = None) -> dict:
    """Read the HK spreadsheet and compose conversion dictionary for HK names.

    The spreadsheet contains the following information:

        - storage mnemonic of the component that generates the HK
        - original HK name (as is comes from the device itself)
        - HK name with the correct prefix
        - name of the column (in the HK file) with the corresponding timestamp

    Args:
        storage_mnemonic: Storage mnemonic of the component for which to compose the conversion dictionary
        use_site: Indicate whether the prefixes of the new HK names are TH-specific
        setup: the Setup to be used, if None, the setup will be loaded from the configuration manager.

    Returns:
        Dictionary with the original HK names as keys and the converted HK names as values.
    """

    setup = setup or load_setup()

    try:
        hk_info_table = setup.telemetry.dictionary
    except AttributeError:
        raise SetupError("Version of the telemetry dictionary not specified in the current setup")

    storage_mnemonic_col = hk_info_table[TmDictionaryColumns.STORAGE_MNEMONIC].values
    correct_name_col = hk_info_table[TmDictionaryColumns.CORRECT_HK_NAMES].values
    original_name_col = hk_info_table[TmDictionaryColumns.ORIGINAL_EGSE_HK_NAMES].values

    selection = np.where(storage_mnemonic_col == storage_mnemonic)

    correct_name_col = correct_name_col[selection]
    original_name_col = original_name_col[selection]

    if use_site:
        th_prefix = f"G{get_site_id()}_"

        th_conversion_dict = {}

        for original_name, correct_name in zip(original_name_col, correct_name_col):
            if str.startswith(str(correct_name), th_prefix):
                th_conversion_dict[original_name] = correct_name

        return th_conversion_dict

    else:
        if len(original_name_col) != len(correct_name_col):
            logger.error(
                f"Name columns in TM dictionary have different length: "
                f"{len(original_name_col)} != {len(correct_name_col)}"
            )

        return dict(zip(original_name_col, correct_name_col))


def get_hk_info(hk_name: str, setup: Optional[Setup] = None) -> tuple:
    """Read the HK spreadsheet and extract information for the given HK parameter.

    The spreadsheet contains the following information:
        - storage mnemonic of the component that generates the HK
        - original HK name
        - HK name with the correct prefix
        - name of the column (in the HK file) with the corresponding timestamp

    Args:
        hk_name: Name of the HK parameter.
        setup: the Setup to use, if None, the setup will be loaded from the configuration manager.

    Returns:
        A tuple where
            (1) the first field contains the storage mnemonic of the component that
            generates the given HK parameter, and
            (2) the second field contains the name of the column in the HK file
            with the corresponding timestamp.

    Raises:
        HKError: when `hk_name` is not known.
    """

    setup = setup or load_setup()
    hk_info_table = setup.telemetry.dictionary

    storage_mnemonic = hk_info_table[TmDictionaryColumns.STORAGE_MNEMONIC].values
    hk_names = hk_info_table[TmDictionaryColumns.CORRECT_HK_NAMES].values
    timestamp_col = hk_info_table[TmDictionaryColumns.TIMESTAMP_NAMES].values

    selection = np.where(hk_names == hk_name)

    try:
        return storage_mnemonic[selection][0], timestamp_col[selection][0]
    except IndexError:
        raise HKError(f"HK parameter {hk_name} unknown")


def get_storage_mnemonics(setup: Setup = None) -> list:
    """Return the list of the storage mnemonics from the TM dictionary.

    Args:
        setup: the Setup to be used, if None, the setup will be loaded from the configuration manager.

    Returns:
        List of the storage mnemonics from the TM dictionary.
    """

    setup = setup or load_setup()

    hk_info_table = setup.telemetry.dictionary
    storage_mnemonics = hk_info_table[TmDictionaryColumns.STORAGE_MNEMONIC].values

    return np.unique(storage_mnemonics).tolist()


def get_housekeeping_names(name_filter: str = None, device_filter: str = None, setup: Setup = None) -> pd.DataFrame:
    """Return HK names, storage mnemonic, and description.

    The TM dictionary is read into a Pandas DataFrame.  If a device filter is given, only the rows pertaining to the
    given storage mnemonic are kept.  If a name filter is given, only keep the rows for which the HK parameter name
    contains the given name filter.

    The result is as a Pandas DataFrame with the following columns:

    - "CAM EGSE mnemonic": Name of the HK parameter;
    - "Storage mnemonic": Storage mnemonic of the device producing the HK;
    - "Description": Description of the HK parameter.

    Synopsis:

    - `get_housekeeping_names(name_filter="RAW", device_filter="N-FEE-HK")`
    - `get_housekeeping_names(name_filter="RAW", device_filter="N-FEE-HK", setup=setup)`

    Args:
        name_filter: Filter the HK dataframe, based on (a part of) the name of the HK parameter(s)
        device_filter: Filter the HK dataframe, based on the given storage mnemonic
        setup: the Setup to be used, if None, the setup will be loaded from the configuration manager.

    Returns:
        Pandas DataFrame with the HK name, storage mnemonic, and description of the HK parameters that pass the
            given filter.
    """

    setup = setup or load_setup()

    hk_info_table = setup.telemetry.dictionary
    hk_info_table.dropna(subset=[TmDictionaryColumns.CORRECT_HK_NAMES], inplace=True)

    if device_filter:
        hk_info_table = hk_info_table.loc[hk_info_table[TmDictionaryColumns.STORAGE_MNEMONIC] == device_filter]

    if name_filter:
        hk_info_table = hk_info_table.query(f'`{TmDictionaryColumns.CORRECT_HK_NAMES}`.str.contains("{name_filter}")')

    return hk_info_table[
        [TmDictionaryColumns.CORRECT_HK_NAMES, TmDictionaryColumns.STORAGE_MNEMONIC, TmDictionaryColumns.DESCRIPTION]
    ]



"""
This module provides functions to calibrate sensor values.
"""

from __future__ import annotations

import numpy as np
from navdict import navdict
from egse.setup import Setup, SetupError


def apply_gain_offset(counts: float, gain: float, offset: float) -> float:
    """Applies the given gain and offset to the given counts.

    Args:
        counts: Uncalibrated, raw data [ADU]
        gain: Gain to apply
        offset: Offset to apply

    Returns:
        Counts after applying the given gain and offset.
    """

    return counts * gain + offset


def counts_to_temperature(sensor_name: str, counts: float, sensor_info: navdict, setup: Setup) -> float | np.ndarray:
    """Converts the given counts for the given sensor to temperature.

    This conversion can be done as follows:

        - (1) Directly from counts to temperature, by applying the gain and offset;
        - (2) Directly from counts to temperature, by applying a function;
        - (3) From counts, via resistance, to temperature.

    Args:
        sensor_name: Sensor name
        counts: Uncalibrated, raw data [ADU]
        sensor_info: Calibration information for the given sensor (type)
        setup: Setup

    Returns:
        Calibrated temperature [°C] for the given sensor
    """

    # (1) Conversion: temperature = counts * gain + offset

    if "counts_to_temperature_gain" in sensor_info and "counts_to_temperature_offset" in sensor_info:
        return apply_gain_offset(
            counts,
            gain=eval(str(sensor_info.counts_to_temperature_gain)),
            offset=sensor_info.counts_to_temperature_offset,
        )

    # (2) Conversion: temperature = func(counts)

    if "counts_to_temperature" in sensor_info:
        # (2a) Polynomial

        if sensor_info.counts_to_temperature.method == "polynomial":
            return np.polyval(sensor_info.counts_to_temperature.counts_to_temperature_coefficients, counts)

    # (3) Conversion: counts -> resistance -> temperature

    else:
        resistance = counts_to_resistance(sensor_name, counts, sensor_info)
        return resistance_to_temperature(sensor_name, resistance, sensor_info, setup)


def counts_to_resistance(sensor_name: str, counts: float, sensor_info: navdict) -> float:
    """Converts the given counts for the given sensor to resistance.

    Args:
        sensor_name: Sensor name
        counts: Uncalibrated, raw data [ADU]
        sensor_info: Calibration information for the given sensor (type)

    Returns:
        Resistance [Ohm] for the given sensor.
    """

    # Offset (if any)

    counts_to_resistance_offset = (
        sensor_info.counts_to_resistance_offset if "counts_to_resistance_offset" in sensor_info else 0
    )

    # Conversion: counts -> voltage -> resistance

    if "counts_to_voltage_gain" in sensor_info and "voltage_to_resistance_gain" in sensor_info:
        return apply_gain_offset(
            counts,
            gain=sensor_info.counts_to_voltage_gain * sensor_info.voltage_to_resistance_gain,
            offset=counts_to_resistance_offset,
        )

    # Conversion: counts -> resistance

    elif "counts_to_resistance_gain" in sensor_info:
        return apply_gain_offset(counts, gain=sensor_info.counts_to_resistance_gain, offset=counts_to_resistance_offset)

    raise SetupError(f"Setup does not contain info for conversion from counts to resistance for {sensor_name}")


def resistance_to_temperature(
    sensor_name: str, resistance: float, sensor_info: navdict, setup: Setup
) -> float | np.ndarray:
    """Converts the given resistance for the given sensor to temperature.

    Args:
        sensor_name: Sensor name
        resistance: Resistance [Ohm]
        sensor_info: Calibration information for the given sensor (type)
        setup: Setup

    Returns:
        Temperature [°C] for the given sensor.
    """

    resistance_to_temperature_info = sensor_info.resistance_to_temperature

    # Series resistance (if any)

    if "series_resistance" in resistance_to_temperature_info:
        series_resistance = resistance_to_temperature_info.series_resistance
        if sensor_name in resistance_to_temperature_info:
            series_resistance = series_resistance[sensor_name]
        resistance -= series_resistance

    method: str = resistance_to_temperature_info.method

    if "divide_resistance_by" in resistance_to_temperature_info:
        resistance /= resistance_to_temperature_info.divide_resistance_by

    # Polynomial

    if method == "polynomial":
        # Coefficients given for conversion temperature -> resistance

        if "temperature_to_resistance_coefficients" in resistance_to_temperature_info:
            return solve_temperature(resistance_to_temperature_info.temperature_to_resistance_coefficients, resistance)

        # Coefficients given for conversion resistance -> temperature

        if "resistance_to_temperature_coefficients" in resistance_to_temperature_info:
            return np.polyval(resistance_to_temperature_info.resistance_to_temperature_coefficients, resistance)

    elif method == "callendar_van_dusen":
        standard = resistance_to_temperature_info.standard
        ref_resistance = resistance_to_temperature_info.ref_resistance

        return callendar_van_dusen(resistance, ref_resistance, standard, setup)

    else:
        raise SetupError(f"Setup does not contain info for conversion from resistance to temperature for {sensor_name}")


def solve_temperature(temperature_to_resistance_coefficients, resistance: float) -> float:
    """Solves the temperature from the temperature -> resistance polynomial.

    For the given temperature -> resistance polynomial and the given resistance, we determine what the corresponding
    temperature is by:

    - Finding the roots of polynomial(temperature) = resistance;
    - Discarding the roots with an imaginary component;
    - Selecting the remaining root in the relevant temperature regime (here: [-200°C, 200°C]).
    """

    temperature_to_resistance_poly = np.poly1d(temperature_to_resistance_coefficients)
    temperatures = (temperature_to_resistance_poly - resistance).roots

    for temperature in temperatures:
        if temperature.imag == 0 and -200 <= temperature <= 200:
            return temperature.real


def callendar_van_dusen(resistance: float, ref_resistance: float, standard: str, setup: Setup) -> float:
    """Solves the Callendar - van Dusen equation for temperature.

    Args:
        resistance: Resistance [Ohm] for which to calculate the temperature
        ref_resistance: Resistance [Ohm] for a temperature of 0°C
        standard: Sensor standard
        setup: Setup

    Returns:
        Temperature [°C] corresponding to the given resistance.
    """

    # Resistances higher than the reference resistance correspond to

    coefficients = setup.sensor_calibration.callendar_van_dusen[standard]

    # Positive temperatures

    if resistance >= ref_resistance:
        resistance_to_temperature_coefficients = [
            ref_resistance * coefficients.C,
            -ref_resistance * 100 * coefficients.C,
            ref_resistance * coefficients.B,
            ref_resistance * coefficients.A,
            ref_resistance * 1,
        ]

    # Negative temperatures

    else:
        resistance_to_temperature_coefficients = [
            ref_resistance * coefficients.B,
            ref_resistance * coefficients.A,
            ref_resistance * 1,
        ]

    return solve_temperature(resistance_to_temperature_coefficients, resistance)


def chebychev(resistance: float, sensor_info: navdict) -> float:
    """Solves the Chebychev equation for temperature.

    Implemented as specified in the calibration certificate of the LakeShore Cernox sensors.

    Args:
        resistance:Resistance [Ohm] for which to calculate the temperature
        sensor_info: Calibration information

    Returns:
        Temperature [°C] corresponding to the given resistance.
    """

    num_fit_ranges = sensor_info.num_fit_ranges

    for fit_range_index in range(1, num_fit_ranges + 1):
        range_info = sensor_info[f"range{fit_range_index}"]

        resistance_lower_limit, resistance_upper_limit = range_info.resistance_range

        if resistance_lower_limit <= resistance <= resistance_upper_limit:
            if range_info.fit_type == "LOG":
                z = np.log10(resistance)

            zl, zu = range_info.z_range
            order = range_info.order
            coefficients = range_info.coefficients

            temperature = 0

            for index in range(0, order + 1):
                k = ((z - zl) - (zu - z)) / (zu - zl)
                temperature += coefficients[index] * np.cos(index * np.arccos(k))

            return temperature


# TODO: Supply voltage calibration
